"""
The RPC(Remote Procedure Call) module.

Handles GET, SET and PUT communication with the server using JSON encoded
messages.

"""
import json
import datetime
import os
import logging


class SeluxitRpc:
    """
    Check connection status.

    Invokes an init_ok() method on reference of connection. If error with
    communication occurs then a connection is closed.

    Args:
        connection: A reference to the socket instance.
        error_msg: Error message while one occurs.

    """

    @staticmethod
    def connection_ok(self, connection, error_msg):
        """
        Check connection status.

        Check the connection status and send an error message if errors have
        occured.

        Args:
            connection: Reference to the connection socket.
            error_msg: Error message to display.

        """
        if not connection.init_ok():
            self.wapp_log.error(error_msg, exc_info=True)
            connection.close()
            quit()

    @staticmethod
    def create_meta(network, network_id):
        """Create meta.

        Creates and returns a meta data about a network.

        Args:
            network:
            network_id: Identify number of a network.

        Returns:
            meta data dictionary.

        """
        return {
            "id": network_id,
            "type": "{}".format(network),
            "version": "2.0"
        }

    @staticmethod
    def get_rpc_success_response(message_id):
        """Get success response.

        Returns a success, JSON formatted, encoded in utf-8 response.

        Args:
            message_id: Id of a message.

        Returns:
            JSON formatted data.

        """
        success_response = {'jsonrpc': '2.0', 'id': message_id, 'result': True}
        return json.dumps(success_response).encode('utf-8')

    @staticmethod
    def get_rpc_fail_response(message_id, text):
        """Get fail response.

        Returns a fail, JSON formatted, encoded in utf-8 response with a error
        message.

        Args:
            message_id: Id of a message.
            text: A message of error description.

        Returns:
            JSON formatted data.

        """
        error_description = {'code': -32020, 'message': text}
        error_response = {
            'jsonrpc': '2.0',
            'id': message_id,
            'error': error_description
        }
        return json.dumps(error_response).encode('utf-8')

    def __init__(self, save_init):
        """
        Initialize the seluxit_rpc class.

        Initializes an object of seluxit_rpc class by passing required
        parameters. While initialization, wapp_log is created, json_id
        attribute is set to 1. The attribute is increased every time
        information from server is retrieved. Moreover an data_json_rpc
        attribute is created. It is a dictionary that will be used while
        getting information from a server.

        Args:
            save_init: Determines whether or not save json data.

        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())
        self.save_init = save_init
        self.json_id = 1
        self.filename = 'Init_json.txt'
        try:
            os.remove(self.filename)
        except OSError:
            pass
        self.data_json_rpc = {
            'jsonrpc': '2.0',
            'id': self.json_id,
            'params': {}
        }

    def get_next_json_id(self):
        """
        Get next json_id.

        Increases json_id attribute by one every time the method is called.

        Returns:
            json_id: A new, incremented value of json_id attribute.

        """
        self.json_id = self.json_id + 1
        return self.json_id

    def get_rpc_network(self, network_id, network_name, put=True):
        """
        Retrieve network from server.

        Retrieves network from a server. The method is used while adding a
        network. Depending on request method it fills data_json_rpc dictionary
        with different information.

        Args:
            network_id: Unique identifying number of a network
            network_name: Name of a network
            put: defines if the request method is put {default: True}

        Returns:
            JSON formatted data of network

        """
        network = "network"
        meta = self.create_meta(network, network_id)
        data_inside = {
            "meta": meta,
            'name': network_name
        }

        if put:
            network_parameters = {
                'url': '/{}/{}'.format(network, network_id),
                'data': data_inside
            }
            self.data_json_rpc['method'] = 'PUT'
        else:
            network_parameters = {
                'url': '/{}'.format(network),
                'data': data_inside
            }
            self.data_json_rpc['method'] = 'POST'

        self.data_json_rpc['params'] = network_parameters
        self.data_json_rpc['id'] = self.get_next_json_id()
        return json.dumps(self.data_json_rpc).encode('utf-8')

    def get_rpc_device(
            self,
            network_id,
            device_id,
            name,
            manufacturer,
            product,
            version,
            serial,
            description,
            protocol,
            communication,
            included,
            put=True
    ):
        """
        Retrieve device from server.

        Retrieves device from a server. The method is used while adding a
        device. It is checked if attributes from meta data are the same as ones
        from objects. Then depending on request method it fills data_json_rpc
        dictionary with different information.

        Args:
            network_id: Unique identifying number of the network.
            device_id: Unique identifying number of the device.
            name: Name of the device.
            manufacturer: The manufacturer of the device.
            product: The product the device belongs to.
            version: The version of the device.
            serial: The device serial number.
            description: The description of the device.
            protocol: The device's protocol.
            communication: The communication type of the device.
            included: Defines if the device is included in the network or not.
            put: Defines if the request is of type PUT. (default: {True})

        Returns:
            JSON formatted data.

        """
        device = "device"
        meta = self.create_meta(device, device_id)
        device_data = {'name': name, "meta": meta, 'included': included}

        if manufacturer:
            device_data['manufacturer'] = manufacturer

        if product:
            device_data['product'] = product

        if version:
            device_data['version'] = version

        if serial:
            device_data['serial'] = serial

        if description:
            device_data['description'] = description

        if protocol:
            device_data['protocol'] = protocol

        if communication:
            device_data['communication'] = communication

        if put:
            self.data_json_rpc['method'] = 'PUT'
            network_parameters = {
                'url': '/network/{}/{}/{}'.format(
                    network_id,
                    device,
                    device_id
                ),
                'data': device_data}
        else:
            self.data_json_rpc['method'] = 'POST'
            network_parameters = {
                'url': '/network/{}/{}'.format(network_id, device),
                'data': device_data
            }

        self.data_json_rpc['params'] = network_parameters
        self.data_json_rpc['id'] = self.get_next_json_id()
        self.wapp_log.info(self.data_json_rpc)
        return json.dumps(self.data_json_rpc).encode('utf-8')

    def base_device_value(
            self,
            name,
            permission,
            value_id,
            special_type,
            base_type
    ):
        """
        Retrieve base value.

        Retrieves base value from a server. Because there
        are three different kinds of values, this method helps to return
        only common attributes between all of them. The method is used while
        getting number, string and blob values.

        Args:
            name: Name of a value.
            permission: Permission of a value
            value_id: Unique identifier of a value
            special_type: Type of value
            base_type: "blob", "string", "number"

        Returns:
            device_value dictionary containing common information about a
            value.

        """
        device_value = {'status': 'ok', 'name': name, 'permission': permission}
        value = "value"
        device_value["meta"] = self.create_meta(value, value_id)

        if not special_type:
            device_value['type'] = name
        else:
            device_value['type'] = special_type

        device_value[base_type] = {}
        return device_value

    def get_rpc_value_string(
            self,
            network_id,
            device_id,
            value_id,
            name,
            specific_type,
            permission,
            max_val_len,
            encoding,
            period,
            delta,
            put=True
    ):
        """
        Retrieve string-type value.

        Retrieves string value from a server. Uses functionality of
        base_device_value to get a directory with common attributes of value
        class.

        Args:
            network_id: Name of a value.
            device_id: Permission of a value.
            value_id: Unique identifier of a value.
            name: Type of value.
            specific_type: "blob", "string", "number".
            permission: Permission of a value.
            max_val_len: Maximum length of a value's string.
            encoding: Value's encoding.
            period: Time after which a value should send report.
            delta: Difference between val1 and val2 over time to check for.
            put: Defines if the request is of type PUT. (default: {True})

        Returns:
            JSON formatted data of string value.

        """
        device_value = self.base_device_value(
            name,
            permission,
            value_id,
            specific_type,
            'string'
        )

        if max_val_len:
            device_value['string']['max'] = max_val_len

        if period:
            device_value['string']['period'] = period

        if delta:
            device_value['string']['delta'] = delta

        if encoding:
            device_value['string']['encoding'] = encoding

        self.data_json_rpc['id'] = self.get_next_json_id()
        self.create_json_message(
            device_id,
            network_id,
            value_id,
            device_value,
            put
        )
        return json.dumps(self.data_json_rpc).encode('utf-8')

    def get_rpc_value_number(
            self,
            network_id,
            device_id,
            value_id,
            name,
            specific_type,
            permission,
            min_val,
            max_val,
            step,
            period,
            delta,
            unit,
            put=True
    ):
        """
        Retrieve number-type value.

        Retrieves number value from a server. Uses functionality of
        base_device_value to get a directory with common attributes of value
        class.

        Args:
            network_id: Name of a value.
            device_id: Permission of a value.
            value_id: Unique identifier of a value.
            name: Type of value.
            specific_type: "blob", "string", "number".
            permission: Permission of a value.
            min_val: Minimum of a value's number.
            max_val: Maximum of a value's number.
            encoding: Value's encoding.
            period: Time after which a value should send report.
            delta: Difference between val1 and val2 over time to check for.
            put: Defines if the request is of type PUT. (default: {True})

        Returns:
            JSON formatted data of a number value.

        """
        device_value = self.base_device_value(
            name,
            permission,
            value_id,
            specific_type,
            'number'
        )
        device_value['number']['min'] = min_val
        device_value['number']['max'] = max_val
        device_value['number']['step'] = step

        if unit:
            device_value['number']['unit'] = unit

        if period:
            device_value['number']['period'] = period

        if delta:
            device_value['number']['delta'] = delta

        self.data_json_rpc['id'] = self.get_next_json_id()
        self.create_json_message(
            device_id,
            network_id,
            value_id,
            device_value,
            put
        )
        return json.dumps(self.data_json_rpc).encode('utf-8')

    def get_rpc_value_blob(
            self,
            network_id,
            device_id,
            value_id,
            name,
            specific_type,
            permission,
            max_val,
            encoding,
            period,
            delta,
            put=True
    ):
        """
        Retrieve blob-type value.

        Retrieves blob value from a server. Uses functionality of
        base_device_value to get a directory with common attributes of value
        class.

        Args:
            network_id: Name of a value.
            device_id: Permission of a value.
            value_id: Unique identifier of a value.
            name: Type of value.
            specific_type: "blob", "string", "number".
            permission: Permission of a value.
            max_val: Maximum of a value's blob.
            encoding: Value's encoding.
            period: Time after which a value should send report.
            delta: Difference between val1 and val2 over time to check for.
            put: Defines if the request is of type PUT. (default: {True})

        Returns:
            JSON formatted data of a blob value.

        """
        device_value = self.base_device_value(
            name,
            permission,
            value_id,
            specific_type,
            'blob'
        )

        if max_val:
            device_value['blob']['max'] = max_val

        if period:
            device_value['blob']['period'] = period

        if delta:
            device_value['blob']['delta'] = delta

        if encoding:
            device_value['blob']['encoding'] = encoding

        self.data_json_rpc['id'] = self.get_next_json_id()
        self.create_json_message(
            device_id,
            network_id,
            value_id,
            device_value,
            put
        )
        return json.dumps(self.data_json_rpc).encode('utf-8')

    def get_rpc_value_set(
            self,
            network_id,
            device_id,
            value_id,
            name,
            specific_type,
            permission,
            element,
            period,
            delta,
            put=True
    ):

        # TODO(Dimitar/Jakub): Describe this later, unsure of what it does.
        # (Possibly ask Karsten when he's back)
        """
        [summary].

        [description]

        Args:
            network_id: [description]
            device_id: [description]
            value_id: [description]
            name: [description]
            specific_type: [description]
            permission: [description]
            element: [description]
            period: [description]
            delta: [description]
            put: [description] (default: {True})

        """
        device_value = self.base_device_value(
            name,
            permission,
            value_id,
            specific_type,
            'set'
        )

        if period:
            device_value['set']['period'] = period
        if element:
            device_value['set']['element'] = element
        if delta:
            device_value['set']['delta'] = delta

        self.data_json_rpc['id'] = self.get_next_json_id()

        self.create_json_message(
            device_id,
            network_id,
            value_id,
            device_value,
            put
        )
        return json.dumps(self.data_json_rpc).encode('utf-8')

    def get_rpc_state(
            self,
            data,
            network_id,
            device_id,
            value_id,
            report_id,
            set_type,
            get=False,
            put=True,
            trace_id=None,
            state_obj=None
    ):
        """
        Retrieve state of the value.

        Retrieves the value state from a server. Uses functionality of
        base_device_value to get a directory with common attributes of value
        class.

        Args:
            data: Data from the state of the value.
            network_id: Unique identifier of the network.
            device_id: Unique identifier of a device.
            value_id: Unique identifier of a value.
            report_id: Unique identifier of the report state.
            set_type: The type to set.
            get: Defines if the request is of type GET. (default: {False})
            put: Defines if the request is of type PUT. (default: {True})
            trace_id: ID of the debug trace. (default: {None})
            state_obj: Reference to the value instance's state instance.
                (default: {None})

        Returns:
            JSON formatted data of the state.

        """
        update = datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%fZ')
        device_state = {
            'meta': {},
            'type': '',
            'status': 'Send',
            'data': data,
            'timestamp': update
        }
        state = "state"
        device_state["meta"] = self.create_meta(state, report_id)

        if set_type == 'report':
            if state_obj is not None:
                state_obj.timestamp = update
            set_type = 'Report'
        else:
            if state_obj is not None:
                state_obj.timestamp = update
            set_type = 'Control'

        device_state['type'] = set_type
        self.data_json_rpc['id'] = self.get_next_json_id()
        if get is True and put is False:
            self.create_json_message(
                device_id,
                network_id,
                value_id,
                device_state,
                put=True,
                state=state,
                state_id=report_id,
                trace_id=trace_id,
                get=True
            )
            return json.dumps(self.data_json_rpc).encode('utf-8')
        self.create_json_message(
            device_id,
            network_id,
            value_id,
            device_state,
            put,
            state=state,
            state_id=report_id,
            trace_id=trace_id
        )
        return json.dumps(self.data_json_rpc).encode('utf-8')

    def get_state_control(
            self,
            connection,
            data,
            network_id,
            device_id,
            value_id,
            control_id,
            get=True
    ):
        """
        Retrieve the control state from the server.

        Gets a value's control state information.Uses the generic
        get_rpc_state() method.

        Args:
            connection: A reference to the socket instance.
            data: Data from the value's control state.
            network_id: Unique identifying number of the network.
            device_id: Unique identifying number of the device.
            value_id: Unique identifying number of the value.
            control_id: The value's control state ID.
            get: Defines if the request is of type GET. (default: {True})

        Returns:
            JSON formatted data if any is retrieved, False otherwise.

        """
        json_data = self.get_rpc_state(
            data,
            network_id,
            device_id,
            value_id,
            control_id,
            "control",
            get=get,
            put=False
        )

        self.send_init_json(connection, json_data)
        self.connection_ok(self, connection, "Error in state control")
        if json_data is not None:
            return json_data
        else:
            return False

    def create_json_message(
            self,
            device_id,
            network_id,
            value_id,
            data,
            put,
            state=None,
            state_id=None,
            trace_id=None,
            get=False
    ):
        """
        Create a JSON encoded message.

        Creates a message that is used for communicating with a server. Both
        for put and post as well as get request methods. The message is build
        as a url and contains information about a network, device, value and
        state.

        Args:
            device_id: Unique identifying number of device.
            natwork_id: Unique identifying number of network.
            value_id: Unique identifying number of value.
            data: Passed data around which a message will be created.
            put: Determines whether or not it is put request.
            state: reference to a state object.
            state_id: Unique identifying number of state.
            trace_id: Id of trace is necessary.
            get: Defines if the request is of type GET. (default: {False})

        """
        network_parameters = {'url': '', 'data': data}
        base_url = '/network/{}/device/{}/value/'.format(
            network_id,
            device_id
        )
        if put:
            if get:
                if state == 'state':
                    network_parameters['url'] = "{}{}/{}/{}".format(
                        base_url,
                        value_id,
                        state,
                        state_id
                    )
                else:
                    network_parameters['url'] = "{}{}".format(
                        base_url,
                        value_id
                    )

                self.data_json_rpc['method'] = 'GET'
                if trace_id:
                    network_parameters['url'] = "{}?trace={}".format(
                        network_parameters['url'],
                        trace_id
                    )
            else:
                if state == 'state':
                    network_parameters['url'] = "{}{}/{}/{}".format(
                        base_url,
                        value_id,
                        state,
                        state_id
                    )
                else:
                    network_parameters['url'] = "{}{}".format(
                        base_url,
                        value_id
                    )
                self.data_json_rpc['method'] = 'PUT'
                if trace_id:
                    network_parameters['url'] = "{}?trace={}".format(
                        network_parameters['url'],
                        trace_id
                    )
        else:
            if state == 'state':
                network_parameters['url'] = "{}{}/{}".format(
                    base_url,
                    value_id,
                    state
                )
            else:
                network_parameters['url'] = base_url
            self.data_json_rpc['method'] = 'POST'

        self.data_json_rpc['params'] = network_parameters

    # Used by initialize
    def add_network(
            self,
            connection,
            network_id,
            network_name
    ):
        """Add an instance of the Network class.

        While initializing adds network to send and receive queue.

        Args:
            connection: A reference to the socket instance.
            network_id: Unique identifying number of the network.
            network_name: Name of the network.

        """
        json_data = self.get_rpc_network(network_id, network_name, put=False)
        self.send_init_json(connection, json_data)
        self.connection_ok(self, connection, "Error in add network")

    def add_device(
            self,
            connection,
            network_id,
            device_id,
            name,
            manufacturer,
            product,
            version,
            serial,
            description,
            protocol,
            communication,
            included
    ):
        """Add an instance of the Device class.

        While initializing adds a device to the send and receive queue.

        Args:
            connection: A reference to the socket instance.
            network_id: Unique identifying number of the network.
            device_id: Unique identifying number of the device.
            name: Name of the device.
            manufacturer: The manufacturer of the device.
            product: The product the device belongs to.
            version: The version of the device.
            serial: The device serial number.
            description: The description of the device.
            protocol: The device's protocol.
            communication: The communication type of the device.
            included: Defines if the device is included in the network or not.

        """
        json_data = self.get_rpc_device(
            network_id,
            device_id,
            name,
            manufacturer,
            product,
            version,
            serial,
            description,
            protocol,
            communication,
            included,
            put=False
        )
        self.send_init_json(connection, json_data)
        self.connection_ok(self, connection, "Error in add device")

    def send_init_json(self, connection, json_data):
        """Send initial JSON data.

        Sends the initial state of the JSON data before instantiation or
        modification of it.

        Args:
            connection: Reference to the connection socket.
            json_data: Initial JSON data.

        """
        connection.send_data(json_data)
        if self.save_init:
            with open(self.filename, 'a+') as file:
                file.write(str(json_data) + "\n")

    def add_value_string(
            self,
            connection,
            network_id,
            device_id,
            value_id,
            name,
            specific_type,
            permission,
            max_val_len,
            encoding,
            period,
            delta
    ):
        """Add an instance of the Value class with a string.

        While initializing adds a value with a string attributes to the send
        and receive queue.

        Args:
            connection: A reference to the socket instance.
            network_id: Unique identifying number of the network.
            device_id: Unique identifying number of the device.
            value_id: Unique identifying number of the value.
            name: Name of the value.
            specific_type: The type of value.
            permission: Permission of a value.
            max_val: Maximum of a value's string.
            encoding: Value's encoding.
            period: Time after which a value should send report.
            delta: Difference between val1 and val2 over time to check for.

        """
        json_data = self.get_rpc_value_string(
            network_id,
            device_id,
            value_id,
            name,
            specific_type,
            permission,
            max_val_len,
            encoding,
            period,
            delta,
            put=False
        )

        self.send_init_json(connection, json_data)
        self.connection_ok(self, connection, "Error in value string")

    def add_value_number(
            self,
            connection,
            network_id,
            device_id,
            value_id,
            name,
            specific_type,
            permission,
            min_val,
            max_val,
            step,
            period,
            delta,
            unit
    ):
        """Add an instance of the Value class with a number.

        While initializing adds a value with a number attributes to the send
        and receive queue.

        Args:
            connection: A reference to the socket instance.
            network_id: Unique identifying number of the network.
            device_id: Unique identifying number of the device.
            value_id: Unique identifying number of the value.
            name: Name of the value.
            specific_type: The type of value.
            permission: Permission of a value.
            min_val: Minimum value.
            max_val: Maximum value.
            step: Step between values.
            period: Time after which a value should send report.
            delta: Difference between val1 and val2 over time to check for.
            unit: Unit of measurement for the value.

        """
        json_data = self.get_rpc_value_number(
            network_id,
            device_id,
            value_id,
            name,
            specific_type,
            permission,
            min_val,
            max_val,
            step,
            period,
            delta,
            unit,
            put=False
        )

        self.send_init_json(connection, json_data)
        self.connection_ok(self, connection, "Error in value number")

    def add_value_blob(
            self,
            connection,
            network_id,
            device_id,
            value_id,
            name,
            specific_type,
            permission,
            max_val,
            encoding,
            period,
            delta
    ):
        """Add an instance of the Value class with a blob.

        While initializing adds a value with a blob attributes to the send and
        receive queue.

        Args:
            connection: A reference to the socket instance.
            network_id: Unique identifying number of the network.
            device_id: Unique identifying number of the device.
            value_id: Unique identifying number of the value.
            name: Name of the value.
            specific_type: The type of value.
            permission: Permission of a value.
            max_val: Maximum of a value's string.
            encoding: Value's encoding.
            period: Time after which a value should send report.
            delta: Difference between val1 and val2 over time to check for.

        """
        json_data = self.get_rpc_value_blob(
            network_id,
            device_id,
            value_id,
            name,
            specific_type,
            permission,
            max_val,
            encoding,
            period,
            delta,
            put=False
        )

        self.send_init_json(connection, json_data)
        self.connection_ok(self, connection, "Error in value blob")

    def add_value_set(
            self,
            connection,
            network_id,
            device_id,
            value_id,
            name,
            specific_type,
            permission,
            element,
            period,
            delta
    ):
        # TODO(Dimitar/Jakub): Describe this later, unsure of what it does.
        # (Possibly ask Karsten when he's back)
        """[summary].

        [description]

        Args:
            connection: [description]
            network_id: [description]
            device_id: [description]
            value_id: [description]
            name: [description]
            specific_type: [description]
            permission: [description]
            element: [description]
            period: [description]
            delta: [description]

        """
        json_data = self.get_rpc_value_set(
            network_id,
            device_id,
            value_id,
            name,
            specific_type,
            permission,
            element,
            period,
            delta,
            put=False
        )

        self.send_init_json(connection, json_data)
        self.connection_ok(self, connection, "Error in value set")

    def add_state_report(
            self,
            connection,
            data,
            network_id,
            device_id,
            value_id,
            report_id
    ):
        """Add an instance of the State class of type Report.

        While initializing adds a State of type Report to the send and
        receive queue.

        Args:
            connection: A reference to the socket instance.
            data: Data from report state.
            network_id: Unique identifying number of the network.
            device_id: Unique identifying number of the device.
            value_id: Unique identifying number of the value.
            report_id: Unique identifying number of the report state.

        """
        json_data = self.get_rpc_state(
            data,
            network_id,
            device_id,
            value_id,
            report_id,
            "report",
            put=False
        )

        self.send_init_json(connection, json_data)
        self.connection_ok(self, connection, "Error in state report")

    def add_state_control(
            self,
            connection,
            data,
            network_id,
            device_id,
            value_id,
            control_id
    ):
        """Add an instance of the State class of type Control.

        While initializing adds a State of type Control to the send and
        receive queue.

        Args:
            connection: A reference to the socket instance.
            data: Data from report state.
            network_id: Unique identifying number of the network.
            device_id: Unique identifying number of the device.
            value_id: Unique identifying number of the value.
            control_id: Unique identifying number of the control state.

        """
        json_data = self.get_rpc_state(
            data,
            network_id,
            device_id,
            value_id,
            control_id,
            "control",
            put=False
        )

        self.send_init_json(connection, json_data)
        self.connection_ok(self, connection, "Error in state control")



"""
Status module for the wappsto module.

Status module to keep track of what the module status is.

Attributes:
    STARTING: Starting phase status.
    INSTANTIATING: Instantiating phase status.
    CONNECTING: Connecting phase status.
    RECONNECTING: Reconnecting phase status.
    CONNECTED: Connected phase status.
    INITIALIZING: Initializing phase status.
    STARTING_THREADS: Thread starting phase status.
    RUNNING: Running phase status.
    DISCONNECTING: Disconnecting phase status.

"""
import logging


STARTING = "Starting"
INSTANTIATING = "Instantiating"
CONNECTING = "Connecting"
RECONNECTING = "Reconnecting"
CONNECTED = "Connected"
INITIALIZING = "Initializing"
STARTING_THREADS = "Starting Threads"
RUNNING = "Running"
DISCONNECTING = "Disconnecting"


class Status:
    """
    Status tracking class.

    Tracks the status of the wappsto module through global variables that are
    changed at runtime.
    """

    def __init__(self):
        """
        Initialize the Status class.

        Initializes the Status class, which handles changing the program's
        status flag.

        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())
        self.callback = self.callback_not_set()
        self.current_status = None

    def callback_not_set(self):
        """
        Message about no callback being set.

        Temporary method to signify that there is no callback set for the
        class.

        Returns:
            "Callback not set" message.

        """
        return self.wapp_log.info("Callback not set.")

    def set_callback(self, callback):
        """
        Set callback reference.

        Sets the reference to the method to the balled whenever the status
        is changed.

        Args:
            callback: Function callback reference.

        """
        self.callback = callback
        self.wapp_log.info("Callback {} has been set.".format(callback))

    def get_status(self):
        """
        Retrieve current status.

        Retrieves the class' current_status attribute.

        Returns:
            Current status attribute.

        """
        return self.current_status

    def set_status(self, status):
        """
        Set the current status.

        Sets the current status from the global attribute states.

        Args:
            status: Status global attribute.

        """
        self.current_status = status
        if self.callback is not None:
            self.callback(self)



"""The object insantiation and saving package.

Handles creating object instances from a parsed JSON file. It also saves the
runtime instances to a JSON file that is similar in structure to the intiially
parsed one using a custom encoding class.
"""



"""
The network module.

Stores attributes for the network instance.
"""
import logging


class Network:
    """
    Network instance class.

    Stores attributes for the network instance.
    """

    def __init__(self, uuid, version, name):
        """
        Initialize the Network class.

        Initializes an object of network class by passing required parameters.

        Args:
            uuid: Unique identifier of a network
            version: Version of a network
            name: Name of a network

        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())
        self.uuid = uuid
        self.version = version
        self.name = name
        msg = "Network {} Debug \n{}".format(name, str(self.__dict__))
        self.wapp_log.debug(msg)



"""
The guider module.

Handles first time guide messages.
"""


class Guider:
    """
    The Guider class.

    Prints the first time guide messages at each step of the program's
    execution and uses input() to block execution while the user reads the
    messages.
    """

    @staticmethod
    def starting_server():
        """
        Socket and RPC (Remote Procedure Call) setup message.

        A message that describes instancing a connection socket to the server
        and the RPC class.
        """
        input("We will try to connect to the server using fixed address and "
              "port. If we will not connect successfully the program will "
              "attemp to reconnect five times. After fifth failure the "
              "program will stop. To continue click [ENTER].")

    @staticmethod
    def instantiating_objects():
        """
        Object instantiation message.

        A message that describes instantiating the network, device, value and
        state objects from the given JSON file.
        """
        input("Now data from json file will be initialized into list "
              "of devices and each of the device will include own list of "
              "values containing their states. Moreover callbacks will be"
              "added to value objects. To continue click [ENTER].")

    @staticmethod
    def connecting_to_server():
        """
        Connect to the server message.

        A message that preceeds the attempt to connect to the server.
        """
        input("Click anything to connect to the server. To continue click"
              " [ENTER].")

    @staticmethod
    def reconnecting_to_server():
        """
        Reconnect attempt message.

        A message that preceeds a reconnection attempt to the server.
        """
        input("Click [ENTER] to start another attempt to connect to the "
              "server.")

    @staticmethod
    def connected_to_server():
        """
        Establish connection message.

        A message to notify the user that they have successfully established a
        connection to the server.
        """
        input("You have successfully connected to the server. To continue "
              "click [ENTER].")

    @staticmethod
    def initializing():
        """
        Initialize object instances on the send and receive queue.

        A message that describes the initialization of the instanced objects by
        adding them to the sending and receiving queue.
        """
        input("The objects instances will be added to sending and "
              "receiving queues by getting their attributes. To continue "
              "click [ENTER].")

    @staticmethod
    def starting_threads():
        """
        Thread start message.

        A message that decribres the start of the sending and receiving
        threads.
        """
        input("Starting sending and receiving threads that will exchange "
              "information with server. To continue click [ENTER].")

    @staticmethod
    def running():
        """
        Thread run message.

        A message that notifies the user that the threads have began running
        and can accept sending and receiving data.
        """
        input("The threads are running. To continue click [ENTER].")

    @staticmethod
    def disconnecting():
        """
        Disconnect message.

        A message that notifies the user that the program is disconnecting from
        the server.
        """
        input("Disconnecting from server and stopping the program. "
              "To continue click [ENTER].")



"""
Send data module.

Sends data packages to the server and stores global state flags.

Attributes:
    SEND_SUCCESS: Successful sending response global flag.
    SEND_FAILED: Failed sending response global flag.
    SEND_REPORT: Report message global flag.
    SEND_RECONNECT: Reconnection attempt global flag.
    SEND_CONTROL: Control message global flag.
    SEND_TRACE: Trace sending global flag.

"""
SEND_SUCCESS = 1
SEND_FAILED = 2
SEND_REPORT = 3
SEND_RECONNECT = 4
SEND_CONTROL = 5
SEND_TRACE = 6


class SendData:
    """
    Send data class.

    Stores data package attributes to send to the server.
    """

    def __init__(
            self,
            msg_id,
            rpc_id=None,
            data=None,
            network_id=None,
            device_id=None,
            value_id=None,
            state_id=None,
            text=None,
            parent=None,
            trace_id=None,
            control_value_id=None
    ):
        """
        Initialize the SendData class.

        Holds data sending values so they can be accessed from the RPC
        and socket.

        Args:
            msg_id: The ID of the message.
            rpc_id: The ID of the RPC instance. (default: {None})
            data: The data to send. (default: {None})
            network_id: The ID of the network to interact with.
            (default: {None})
            device_id: The ID of the device to interact with. (default: {None})
            value_id: The ID of the value to interact with. (default: {None})
            state_id: The ID of the state to interact with. (default: {None})
            text: The message to send. (default: {None})
            parent: The parent object. (default: {None})
            trace_id: The trace ID used to debug. (default: {None})
            control_value_id: The control data ID (default: {None})

        """
        self.msg_id = msg_id
        self.rpc_id = rpc_id
        self.data = data
        self.network_id = network_id
        self.device_id = device_id
        self.value_id = value_id
        self.state_id = state_id
        self.text = text
        self.parent = parent
        self.trace_id = trace_id
        self.control_value_id = control_value_id



"""
The raspberrypi object instantiation module.

Represents raspberrypi components from JSON to class instances.
"""
import json
from json import JSONDecodeError
import logging
from . import save_objects
from ..connection.network_classes import network
from ..connection.network_classes import device
from ..connection.network_classes import value
from ..connection.network_classes import state
from . import status
from ..user_guide.guider import Guider


class Instantiator:
    """
    Create class instances from JSON file.

    Creates instances of network, devices, values and states by parsing a
    given JSON file.
    """

    def __init__(
            self,
            json_file_name,
            load_from_state_file,
            status,
            guide,
            path_to_calling_file):
        """
        Initialize the Instantiator class.

        Initializes the Instatiator class which creates and holds instances of
        the network, devices, values and their states.

        Args:
            json_file_name: The name of the JSON file to parse.
            load_from_state_files: A True/False flag to denote whether to load
                from the saved files directory.
            status: Reference to the Status instance to update the program's
                status flag.
            guide: Reference to the Guider instance to update the first time
                guide.

        Raises:
            JSONDecodeError: Exception when trying to parse the JSON file.
            FileNotFoundError: Exception when trying to find the file's
                location.

        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())
        self.path_to_calling_file = path_to_calling_file
        self.network_cl = None
        self.device_list = []
        self.status = status
        self.load_from_state_file = load_from_state_file
        self.guide = guide
        self.json_file_name = json_file_name

        if self.load_from_state_file:
            object_saver = save_objects.SaveObjects(path_to_calling_file)
            self.json_file_name = object_saver.load_instance()
        try:
            with open(self.json_file_name) as data_file:
                self.decoded = json.loads(data_file.read())
            self.parse_json_file()
            msg = "Classes instantiated from: {}".format(json_file_name)
            self.wapp_log.debug(msg)
        except FileNotFoundError as fnfe:
            self.wapp_log.error("Error finding file: {}".format(fnfe))
            raise fnfe

    def parse_json_file(self):
        # TODO(Dimitar): Fill in exception
        """
        Parse JSON file.

        Opens the JSON file from the location attribute and parses the data
        inside of it to a class attribute.

        Raises:
            Exception: TO BE FILLED IN.

        """
        self.wapp_log.debug("Opening file: {}".format(self.json_file_name))

        try:
            self.json_container = json.loads(self.decoded.get('data'))
        except JSONDecodeError as jde:
            self.wapp_log.error("Error decoding: {}".format(jde))
            raise jde

        self.wapp_log.debug("RAW JSON DATA:\n\n{}\n\n".format(
            self.json_container)
        )
        self.network_cl = self.build_network(self.decoded)
        self.status.set_status(status.INSTANTIATING)
        if self.guide:
            Guider.instantiating_objects()
        self.device_list = self.build_device_list()

    def build_device_list(self):
        # TODO(Dimitar): Fill in exception
        """
        Build list of device instances.

        Builds a list of instances of devices that contain instances of values.
        The instances of values also contain instances of states. All are built
        from the decoded JSON file.

        Returns:
            A list of devices instances.

        """
        for device_iterator in self.json_container.get('device'):
            uuid = device_iterator.get('meta').get('id')
            name = device_iterator.get('name')
            product = device_iterator.get('product')
            protocol = device_iterator.get('protocol')
            serial_number = device_iterator.get('serial')
            version = device_iterator.get('meta').get('version')
            manufacturer = device_iterator.get('manufacturer')
            communication = device_iterator.get('communication')
            description = device_iterator.get('description')

            # CHANGE THIS LATER
            device_cl = device.Device(
                parent_network=self.network_cl,
                uuid=uuid,
                name=name,
                product=product,
                protocol=protocol,
                serial_number=serial_number,
                version=version,
                manufacturer=manufacturer,
                communication=communication,
                description=description
            )
            self.device_list.append(device_cl)
            self.wapp_log.debug("Device {} appended to {}"
                                .format(device_cl, self.device_list)
                                )

            for value_iterator in device_iterator.get('value'):
                uuid = value_iterator.get('meta').get('id')
                name = value_iterator.get('name')
                type_of_value = value_iterator.get('type')
                permission = value_iterator.get('permission')
                init_value = value_iterator.get('state')[0].get('data')
                data_type = None
                number_min = None
                number_max = None
                number_step = None
                number_unit = None
                string_encoding = None
                string_max = None
                blob_encoding = None
                blob_max = None

                if 'string' in value_iterator:
                    data_type = 'string'
                    get_string = value_iterator.get('string')
                    string_encoding = get_string.get('encoding')
                    string_max = value_iterator.get('string').get('max')
                elif 'blob' in value_iterator:
                    data_type = 'blob'
                    get_bolb = value_iterator.get('blob')
                    blob_encoding = get_bolb.get('encoding')

                    blob_max = value_iterator.get('blob').get('max')
                elif 'number' in value_iterator:
                    data_type = 'number'
                    get_number = value_iterator.get('number')
                    number_min = get_number.get('min')
                    number_max = get_number.get('max')
                    number_step = get_number.get('step')
                    number_unit = get_number.get('unit')

                # CHANGE THIS LATER
                value_cl = value.Value(
                    parent_device=device_cl,
                    uuid=uuid,
                    name=name,
                    type_of_value=type_of_value,
                    data_type=data_type,
                    permission=permission,
                    init_value=init_value,
                    number_max=number_max,
                    number_min=number_min,
                    number_step=number_step,
                    number_unit=number_unit,
                    string_encoding=string_encoding,
                    string_max=string_max,
                    blob_encoding=blob_encoding,
                    blob_max=blob_max
                )

                device_cl.add_value(value_cl)
                self.wapp_log.debug("Value {} appended to {}"
                                    .format(value_cl, device_cl.value_list)
                                    )

                for state_iterator in value_iterator.get('state'):
                    uuid = state_iterator.get('meta').get('id')
                    state_type = state_iterator.get('type')
                    # data = state_iterator.get('data')
                    timestamp = state_iterator.get('timestamp')
                    state_cl = state.State(
                        parent_value=value_cl,
                        uuid=uuid,
                        state_type=state_type,
                        timestamp=timestamp
                    )
                    if state_type == 'Report':
                        value_cl.add_report_state(state_cl)
                        msg = "Report state {} appended to {}"
                        self.wapp_log.debug(msg.format(state_cl, value_cl))
                    elif state_type == 'Control':
                        value_cl.add_control_state(state_cl)
                        msg = "Control state {} appended to {}"
                        self.wapp_log.debug(msg.format(state_cl, value_cl))

        return self.device_list

    def build_network(self, decoded):
        """
        Create network instance.

        Builds a Network class instance by setting the attributes from the
        decoded JSON file.

        Args:
            decoded: Decoded JSON message data.

        Returns:
            A Network class instance.

        """
        decoded_data = json.loads(decoded.get('data'))
        decoded_meta = json.loads(decoded.get('data')).get('meta')
        self.uuid = decoded_meta.get('id')
        self.version = decoded_meta.get('version')
        self.name = decoded_data.get('name')

        network_cl = network.Network(
            uuid=self.uuid,
            version=self.version,
            name=self.name
        )

        self.wapp_log.debug("Network {} built.".format(network_cl))

        return network_cl



"""
The object instance initialization module.

Stores the initialization functionality for the object instances of the
network, devices, values and states.
"""
import logging


class Initialize:
    """
    The initialize class handles initializing object instances.

    Using references to the RPC, socket and object instances, it sends
    activation messages to the server and enables sending/receiving for the
    network, device, value and state instances.
    """

    def __init__(self, rpc):
        """
        Create an initialization instance.

        The initialize instance contains the initialize_all method that is used
        to activate the object instances so that they can send/receive to the
        server.

        Args:
            rpc: Reference to the RPC instance.

        """
        self.rpc = rpc

    def initialize_all(self, conn, instance):
        """
        Initialize all of the devices on the sending and receiving queue.

        Adds all the devices to the send receive queue by pinging the server
        with their information in order to initialize them.

        Args:
            conn: A reference to the socket instance.
            instance: A reference to the network and device instances.

        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())
        self.network_UUID = instance.network_cl.uuid
        self.network_name = instance.network_cl.name

        # Adds the network to the queue.
        self.rpc.add_network(conn, self.network_UUID, self.network_name)
        msg = "Network {} added to Sending queue {}.".format(
            self.network_name,
            self.rpc
        )
        # All of the debug logs may need relocation to the RPC class.
        self.wapp_log.debug(msg)

        for device in instance.device_list:

            self.device_uuid = device.uuid
            self.device_name = device.name
            self.device_manufacturer = device.manufacturer
            self.device_product = device.product
            self.device_version = device.version
            self.device_serial_number = device.serial_number
            self.device_description = device.description
            self.device_protocol = device.protocol
            self.device_communication = device.communication

            # Passes the attributes of the device to the queue add method.
            self.rpc.add_device(
                conn,
                self.network_UUID,
                self.device_uuid,
                self.device_name,
                self.device_manufacturer,
                self.device_product,
                self.device_version,
                self.device_serial_number,
                self.device_description,
                self.device_protocol,
                self.device_communication,
                "1"
            )

            msg = "Device {} added to sending queue {}".format(
                self.device_name,
                self.rpc
            )
            self.wapp_log.debug(msg)

            for value in device.value_list:

                self.value_uuid = value.uuid
                self.value_name = value.name
                self.value_type_of_value = value.type_of_value
                self.value_data_type = value.data_type
                self.value_permission = value.permission
                self.value_init_value = value.init_value or ""
                self.value_blob_encoding = value.blob_encoding
                self.value_blob_max = value.blob_max
                self.value_string_encoding = value.string_encoding
                self.value_string_max = value.string_max
                self.value_number_max = value.number_max
                self.value_number_min = value.number_min
                self.value_number_step = value.number_step
                self.value_number_unit = value.number_unit

                # Passes the attributes of the value to the queue add method.
                # Checks the type of method to determine which variables.
                if self.value_data_type == 'blob':
                    self.rpc.add_value_blob(
                        conn,
                        self.network_UUID,
                        self.device_uuid,
                        self.value_uuid,
                        self.value_name,
                        self.value_type_of_value,
                        self.value_permission,
                        self.value_blob_max,
                        self.value_blob_encoding,
                        "",
                        ""
                    )
                    msg = self.create_message()
                    self.wapp_log.debug(msg)

                elif self.value_data_type == 'string':
                    self.rpc.add_value_string(
                        conn,
                        self.network_UUID,
                        self.device_uuid,
                        self.value_uuid,
                        self.value_name,
                        self.value_type_of_value,
                        self.value_permission,
                        self.value_string_max,
                        self.value_string_encoding,
                        "",
                        ""
                    )

                    msg = self.create_message()
                    self.wapp_log.debug(msg)

                elif self.value_data_type == 'number':
                    self.rpc.add_value_number(
                        conn,
                        self.network_UUID,
                        self.device_uuid,
                        self.value_uuid,
                        self.value_name,
                        self.value_type_of_value,
                        self.value_permission,
                        self.value_number_max,
                        self.value_number_min,
                        self.value_number_step,
                        "",
                        "",
                        self.value_number_unit
                    )

                    msg = self.create_message()
                    self.wapp_log.debug(msg)

                # Adds the states to the respective control or report queues.

                if value.report_state is not None:
                    self.state_type = value.report_state.state_type
                    self.state_uuid = value.report_state.uuid
                    self.rpc.add_state_report(
                        conn,
                        self.value_init_value,
                        self.network_UUID,
                        self.device_uuid,
                        self.value_uuid,
                        self.state_uuid
                    )

                if value.control_state is not None:
                    self.state_type = value.control_state.state_type
                    self.state_uuid = value.control_state.uuid
                    data = self.rpc.get_state_control(
                        conn,
                        self.value_init_value,
                        self.network_UUID,
                        self.device_uuid,
                        self.value_uuid,
                        self.state_uuid
                    )

                    if value.control_state.uuid in str(data):
                        self.rpc.add_state_control(
                            conn,
                            self.value_init_value,
                            self.network_UUID,
                            self.device_uuid,
                            self.value_uuid,
                            self.state_uuid
                        )

    def create_message(self):
        """
        Create a message.

        Returns a message informing about adding a value to a sending queue

        Returns:
            Formatted string message

        """
        return "Value ({}) {} added to sending queue {}.".format(
            self.value_data_type,
            self.value_name,
            self.rpc
        )



"""
The wappsto Encoding module.

Handles encoding object instances to a JSON file for the purpose of saving
them and any modifications made to them.
"""
import logging


class WappstoEncoder:
    """
    The wappsto encoding class.

    Handles encoding the current runtime object instances into JSON. This
    allows the system to be saved as a parsable JSON file similar to the one
    used to start the package.
    """

    def __init__(self):
        """
        Initialize WappstoEncoding.

        Initializes the WappstoEncoding class, which handles encoding the
        various parts of the system into a JSON file.
        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())

    def encode(self, instance):
        """
        Encode instance.

        Encodes objects from the runtime instances and returns the result so
        it can be saved.

        Args:
            instance: Reference to the instance class that holds the object
                instances.

        Returns:
            An encoded JSON result.

        """
        encoded_result = {}
        encoded_result.update(self.encode_network(instance.network_cl))

        for device in instance.device_list:
            encoded_device = self.encode_device(device)
            encoded_result['device'].append(encoded_device)

        return encoded_result

    def encode_network(self, network):
        """
        Encode instance of Network class.

        Handles the condoing of the network instance, contains a template to
        encode the network with.

        Args:
            network: Reference to the instance of the Network class.

        Returns:
            The encoded JSON result.

        """
        encoded_network = {
            "meta": {
                "type": "network",
                "version": "2.0",
                "id": ""
            },
            "name": "",
            "device": []
        }
        encoded_network['meta']['id'] = network.uuid
        encoded_network['name'] = network.name
        self.wapp_log.debug("Network JSON: {}".format(encoded_network))

        return encoded_network

    def encode_device(self, device):
        """
        Encode instance of Device class.

        Handles the encondoing of the device instance, contains a template to
        encode the device with.

        Args:
            device: Reference to the instance of the Device class.

        Returns:
            The encoded JSON result.

        """
        encoded_device = {
            "meta": {
                "type": "device",
                "version": "2.0",
                "id": ""
            },
            "name": "",
            "product": "",
            "serial": "",
            "description": "",
            "protocol": "",
            "communication": "",
            "version": "",
            "manufacturer": "",
            "value": []
        }

        encoded_device['meta']['id'] = device.uuid
        encoded_device['name'] = device.name
        encoded_device['product'] = device.product
        encoded_device['protocol'] = device.protocol
        encoded_device['serial'] = device.serial_number
        encoded_device['meta']['version'] = device.version
        encoded_device['manufacturer'] = device.manufacturer
        encoded_device['communication'] = device.communication
        encoded_device['description'] = device.description

        for value in device.value_list:
            encoded_value = self.encode_value(value)
            encoded_device['value'].append(encoded_value)
        self.wapp_log.debug("Device JSON: {}".format(encoded_device))

        return encoded_device

    def encode_value(self, value):
        """
        Encode instance of Value class.

        Handles the encodoing of the value instance, contains a template to
        encode the value with.

        Args:
            value: Reference to the instance of the Value class.

        Returns:
            The encoded JSON result.

        """
        encoded_value = {
            "meta": {
                "type": "value",
                "version": "2.0",
                "id": ""
            },
            "name": "",
            "permission": "",
            "type": ""
        }

        encoded_value['meta']['id'] = value.uuid
        encoded_value['name'] = value.name
        encoded_value['type'] = value.type_of_value
        encoded_value['permission'] = value.permission

        if value.data_type == 'string':
            encoded_value.update({"string": {}, "state": []})
            if (value.string_encoding is not None
                    and value.string_max is not None):
                encoded_value['string'].update({"encoding": "", "max": "1"})
                encoded_value['string']['encoding'] = value.string_encoding
                encoded_value['string']['max'] = value.string_max
            # TODO (Dimitar): Remove later if unneeded
            # MAY BE NEEDED, LEAVING HERE FOR NOW
            # elif value.string_max is None:
            #     encoded_value['string'].update({"encoding": ""})
            #     encoded_value['string']['encoding'] = value.string_encoding
            # elif value.string_encoding is None:
            #     encoded_value['string'].update({"max": ""})
            #     encoded_value['string']['max'] = value.string_max
        elif value.data_type == 'blob':
            encoded_value.update({"blob": {}, "state": []})
            if value.blob_encoding is not None and value.blob_max is not None:
                encoded_value['blob'].update({"encoding": "", "max": "1"})
                encoded_value['blob']['encoding'] = value.blob_encoding
                encoded_value['blob']['max'] = value.blob_max
            # TODO (Dimitar): Remove later if unneeded
            # MAY BE NEEDED, LEAVING HERE FOR NOW
            # elif value.blob_max is None:
            #     encoded_value['blob'].update({"encoding": ""})
            #     encoded_value['blob']['encoding'] = value.blob_encoding
            # elif value.blob_encoding is None:
            #     encoded_value['blob'].update({"max": ""})
            #     encoded_value['blob']['max'] = value.blob_max
        elif value.data_type == 'number':
            encoded_value.update({"number": {}, "state": []})
            if (
                value.number_min is not None
                and value.number_max is not None
                and value.number_step is not None
                and value.number_unit is not None
            ):
                encoded_value.update({
                    'number': {
                        "min": "",
                        "max": "1",
                        "step": "",
                        "unit": ""
                    },
                    "state": []
                })
                encoded_value['number']['unit'] = value.number_unit

            # TODO (Dimitar): Remove later if unneeded
            # MAY BE NEEDED, LEAVING HERE FOR NOW
            # else:
            #     encoded_value.update({
            #         'number': {
            #             "min": "",
            #             "max": "1",
            #             "step": ""
            #         },
            #         "state": []
            #     })
            encoded_value['number']['min'] = value.number_min
            encoded_value['number']['max'] = value.number_max
            encoded_value['number']['step'] = value.number_step

        encoded_state = None
        if value.report_state:
            encoded_state = self.encode_state(
                value.report_state,
                value.init_value
            )
            encoded_value['state'].append(encoded_state)
        if value.control_state:
            encoded_state = self.encode_state(
                value.control_state,
                value.init_value
            )
            encoded_value['state'].append(encoded_state)
        self.wapp_log.debug("Value JSON: {}".format(encoded_value))

        return encoded_value

    def encode_state(self, state, init_value):
        """
        Encode instance of State class.

        Handles the encodoing of the value instance, contains a template to
        encode the value with.

        Args:
            state: Reference to the instance of the State class.
            init_value: The data in the State.

        Returns:
            The encoded JSON result.

        """
        encoded_state = {
            "meta": {
                "type": "state",
                "version": "2.0",
                "id": "",
                "contract": []
            },
            "data": "",
            "type": "",
            "timestamp": ""
        }
        encoded_state['meta']['id'] = state.uuid
        encoded_state['data'] = init_value
        encoded_state['type'] = state.state_type
        encoded_state['timestamp'] = state.timestamp
        self.wapp_log.debug("State JSON: {}".format(encoded_state))

        return encoded_state



"""
The value module.

Stores attributes for the value instance and handles value-related
methods.
"""
import logging
import threading
import time
import datetime
import re
from math import fabs
from .errors import wappsto_errors


class Value:
    """
    Value instance.

    Stores attributes for the value instance and handles value-related
    methods.
    """

    def __init__(
        self,
        parent_device,
        uuid,
        name,
        type_of_value,
        data_type,
        permission,
        init_value,
        number_max,
        number_min,
        number_step,
        number_unit,
        string_encoding,
        string_max,
        blob_encoding,
        blob_max,
        period=None,
        delta=None
    ):
        """
        Initialize the Value class.

        Initializes an object of value class by passing required parameters.

        Args:
            parent_device: Reference to a device object
            uuid: An unique identifier of a device
            name: A name of a device
            type_of_value: Determines a type of value [e.g temperature, CO2]
            data_type: Defines whether a value is string, blob or number
            permission: Defines permission [read, write, read and write]
            init_value: Initial value after creation of an object
            (if data_type is number then these parameters are relevant):
            number_max: Maximum number a value can have
            number_min: Minimum number a value can have
            number_step: Number defining a step
            number_unit: Unit in which a value should be read
            (if data_type is string then these parameters are relevant):
            string_encoding: A string encoding of a value
            string_max: Maximum length of string
            (if data_type is blob then these parameters are relevant):
            blob_encoding: A blob encoding of a value
            blob_max: Maximum length of a blob

            period: defines the time after which a value should send report
                message. Default: {None})
            delta: defines the a difference of value (default: {None})

        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())
        self.parent_device = parent_device
        self.uuid = uuid
        self.name = name
        self.type_of_value = type_of_value
        self.data_type = data_type
        self.permission = permission
        # The value shared between state instances.
        self.init_value = init_value
        self.number_max = number_max
        self.number_min = number_min
        self.number_step = number_step
        self.number_unit = number_unit
        self.string_encoding = string_encoding
        self.string_max = string_max
        self.blob_encoding = blob_encoding
        self.blob_max = blob_max
        self.period = period
        self.delta = delta
        self.report_state = None
        self.control_state = None
        self.callback = self.callback_not_set()
        self.last_update_of_report = self.get_now()
        self.rpc = None
        self.conn = None
        self.reporting_thread = None
        self.last_update_of_control = None
        self.data_value = init_value
        self.difference = 0
        self.delta_report = 0
        self.last_controlled = self.data_value
        self.parent_network_id = parent_device.get_parent_network().uuid
        self.parent_device_id = parent_device.uuid
        msg = "Value {} debug: {}".format(name, str(self.__dict__))
        self.wapp_log.debug(msg)

    def set_period(self, period):
        """
        Set the value reporting period.

        Sets the period to report a value to the server and starts a thread to
        do so.

        Args:
            period: Reporting period.

        """
        try:
            if self.get_report_state() is not None and int(period) > 0:
                self.period = period

            else:
                self.wapp_log.error("Cannot set the period for this value.")
        except ValueError:
            self.wapp_log.error("Period value must be a number.")

    def set_delta(self, delta):
        """
        Set the delta to report between.

        Sets the delta (range) of change to report in. When a change happens
        in the range of this delta it will be reported.

        Args:
            delta: Range to report between.

        """
        try:
            if (self.is_number_type()
                    and self.get_report_state()
                    and float(delta) > 0):
                self.delta = delta
            else:
                self.wapp_log.error("Cannot set the delta for this value.")
        except ValueError:
            self.wapp_log.error("Delta value must be a number")

    def callback_not_set(self):
        """
        Message about no callback being set.

        Temporary method to signify that there is no callback set for the
        class.

        Returns:
            "Callback not set" message.

        """
        msg = "Callback for value {} was not set".format(self.name)
        return self.wapp_log.info(msg)

    def get_parent_device(self):
        """
        Retrieve parent device reference.

        Gets a reference to the device that owns this device.

        Returns:
            Reference to instance of Device class that owns this Value.

        """
        return self.parent_device

    def add_report_state(self, state):
        """
        Set report state reference to the value list.

        Adds a report state reference to the Value class.

        Args:
            state: Reference to instance of State class.

        """
        self.report_state = state
        msg = "Report state {} has been added.".format(state)
        self.reporting_thread = threading.Thread(target=self.send_report)
        self.reporting_thread.setDaemon(True)
        self.reporting_thread.start()
        self.wapp_log.debug(msg)

    def add_control_state(self, state):
        """
        Set control state reference to the value list.

        Adds a control state reference to the Value class.

        Args:
            state: Reference to instance of State class.

        """
        self.control_state = state
        msg = "Control state {} has been added".format(state)
        self.wapp_log.debug(msg)

    def get_report_state(self):
        """
        Retrieve child report state reference.

        Gets a reference to the child State class.

        Returns:
            Reference to instance of State class.

        """
        if self.report_state is not None:
            return self.report_state
        else:
            msg = "Value {} has no report state.".format(self.name)
            self.wapp_log.warning(msg)

    def get_control_state(self):
        """
        Retrieve child control state reference.

        Gets a reference to the child State class.

        Returns:
            Reference to instance of State class.

        """
        if self.control_state is not None:
            return self.control_state
        else:
            msg = "Value {}  has no control state.".format(self.name)
            self.wapp_log.warning(msg)

    def send_report_delta(self, state):
        """
        Send report message when delta range reached.

        Sends a report message with the current value when the delta range is
        reached.

        Args:
            state: Reference to the report state

        """
        try:
            result = int(self.difference) >= int(self.delta)
            if result and self.rpc is not None and self.delta_report == 1:
                self.send_logic(state, "report")
                self.wapp_log.info("Sent report [DELTA].")
                self.delta_report = 0
                return True
        except AttributeError:
            pass

    def get_now(self):
        """
        Retrieve current time.

        Using datetime library returns current time.

        Returns:
            Current time in format [%Y-%m-%dT%H:%M:%S.%fZ].

        """
        return datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%fZ')

    def send_report(self):
        """
        Send report message.

        Sends a report message with the current value to the server.
        Unless state exists, runs a while loop. It is determined whether or
        not set flag allowing send report because of delta attribute. If delta
        exists: send_report_delta method is called. If period exists it is
        checked if the sum of period value and last time the value was
        updated is greater than current time. If it does then a report is sent.
        Method is running on separate thread which after each loop sleeps for
        one second.
        """
        state = self.get_report_state()
        value = self.last_controlled
        if state is not None:
            while True:
                if self.last_controlled is not None and self.is_number_type():
                    value_check = self.last_controlled
                    if value != value_check:
                        self.difference = fabs(int(value) - int(value_check))
                        value = value_check
                        self.delta_report = 1
                if self.delta is not None:
                    self.send_report_delta(state)
                if self.period is not None:
                    try:
                        last_update_timestamp = self.date_converter(
                            self.last_update_of_report
                        )
                        now = self.get_now()
                        now_timestamp = self.date_converter(now)
                        the_time = last_update_timestamp + self.period
                        if the_time <= now_timestamp and self.rpc is not None:
                            self.wapp_log.info("Sending report [PERIOD].")
                            self.send_logic(state, 'report')
                    except Exception as e:
                        self.reporting_thread.join()
                        self.wapp_log.error(e)
                time.sleep(1)

    def set_callback(self, callback):
        """
        Set the callback.

        Sets the callback attribute. It will be called by the send_logic
        method.

        Args:
            callback: Callback reference.

        Raises:
            CallbackNotCallableException: Custom exception to signify invalid
            callback.

        """
        try:
            if not callable(callback):
                msg = "Callback method should be a method"
                raise wappsto_errors.CallbackNotCallableException(msg)
            self.callback = callback
            self.wapp_log.info("Callback {} has been set.".format(callback))
            return True
        except wappsto_errors.CallbackNotCallableException as e:
            self.wapp_log.error("Error setting callback: {}".format(e))
            raise

    def date_converter(self, date):
        """
        Convert date to timestamp.

        Converts passed date to a timestamp, first removed Z and T charts
        from the date, then using functionality of time and datetime
        libraries, changes the date into timestamp and returns it.

        Args:
            date: string format date

        Returns:
            timestamp
            integer

        """
        date_first = re.sub("Z", "", re.sub("T", " ", date))
        date_format = '%Y-%m-%d %H:%M:%S.%f'
        date_datetime = datetime.datetime.strptime(date_first, date_format)
        return time.mktime(date_datetime.timetuple())

    def send_control(self, data_value):
        """
        Validate before sending a control message.

        Validates the type and permission of a value object and an incoming
        value before sending a control message.

        Args:
            data_value: A new value of the class

        Returns:
            True if successfully, False if error occurs.
            boolean

        Raises:
            ValueError: If a data_value is in incorrect format.

        """
        state = self.get_control_state()
        if state is not None:
            if self.is_number_type():
                try:
                    if self.number_min <= int(data_value) <= self.number_max:
                        return self.send_logic(
                            state,
                            'control',
                            data_value=data_value
                        )
                    else:
                        msg = "Invalid number. Range: {}-{}. Your: {}".format(
                            self.number_min,
                            self.number_max,
                            data_value
                        )
                        self.wapp_log.error(msg)
                        return False
                except ValueError:
                    msg = "Invalid type of value. Must be a number."
                    self.wapp_log.error(msg)
                    return False
            elif not self.is_number_type():
                if self.string_max is None and self.blob_max is None:
                    return self.send_logic(
                        state,
                        'control',
                        data_value=data_value
                    )
                else:
                    if self.data_type == "blob":
                        if len(str(data_value)) <= int(self.blob_max):
                            return self.send_logic(
                                state,
                                'control',
                                data_value=data_value
                            )

                        else:
                            msg = ("Value {} not in correct range for {}"
                                   .format(data_value, self.name))
                            self.wapp_log.error(msg)
                        return False
                    elif self.data_type == "string":
                        if len(str(data_value)) <= int(self.string_max):
                            return self.send_logic(
                                state,
                                'control',
                                data_value=data_value
                            )

                        else:
                            msg = ("Value {} not in correct range for {}"
                                   .format(data_value, self.name))
                            self.wapp_log.error(msg)
                        return False
            else:
                self.wapp_log.error("Device error.")
                return False
        else:
            self.wapp_log.error("Value is read only.")
            return False

    def send_logic(self, state, type, data_value=None):
        """
        Send control or report to a server.

        Sends a control or report message with a new value to the server.
        Updates the time of lasted sent control and sets a new updated value.
        Then invokes callback in order to inform a user about the result of the
        operation.

        Args:
            state: Reference to an instance of a State class.
            type: Determines if the message is report or control
            data_value: A new incoming value.


        Returns:
            True if successfully, False if error occurs.
            boolean

        Raises:
            Exception: If one occurs while sending control message.

        """
        try:
            json_data = self.rpc.get_rpc_state(
                str(data_value),
                self.parent_network_id,
                self.parent_device_id,
                self.uuid,
                state.uuid,
                type,
                state_obj=state
            )
            self.rpc.send_init_json(self.conn, json_data)
            if type == 'report':
                self.last_update_of_report = state.timestamp
                if self.callback is not None:
                    self.callback(self, 'report')
                return True
            else:
                self.data_value = data_value
                self.last_update_of_control = state.timestamp
                self.last_controlled = data_value
                if self.callback is not None:
                    self.callback(self, 'control')
                return True

        except Exception as e:
            msg = "Error reporting state: {}".format(e)
            self.wapp_log.error(msg)
            return False

    def is_number_type(self):
        """
        Validate data type.

        Checks whether the type of the value is number.

        Returns:
        True if the type is number otherwise false.
        boolean

        """
        if self.data_type == "number":
            return True
        else:
            return False



"""
The state module.

Stores attributes for the state instance and handles device-related
"""
import logging


class State:
    """
    State instance class.

    Stores attributes for the state instance and handles device-related
    """

    def __init__(self, parent_value, uuid, state_type, timestamp):
        """
        Initialize the State class.

        Initializes an object of network class by passing required parameters.

        Args:
            parent_value: reference to a value object to which a state
                belongs to
            uuid: unique identifier of a state
            state_type: determines if the state is report or control
            timestamp: time of last update

        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())
        self.parent_value = parent_value
        self.uuid = uuid
        self.state_type = state_type
        self.timestamp = timestamp
        msg = "State {} Debug: \n{}".format(uuid, str(self.__dict__))
        self.wapp_log.debug(msg)



"""
Receive and send handling module.

Handles building sending and receiving data messages and adding them to the
sending queue.
"""
import random
import logging
from . import send_data


def send_trace(sending_queue, parent, trace_id, data, control_value_id=None):
    """
    Add a trace to the sending queue.

    Adds a trace URL to the sending queue for debugging purposes.

    Args:
        sending_queue: The queue requests are being added to.
        parent: Owner of the trace URL.
        trace_id: ID of the trace message.
        data: Trace message data.
        control_value_id: ID of the control state of the value
            (default: {None})

    """
    trace = send_data.SendData(
        send_data.SEND_TRACE,
        parent=parent,
        trace_id=trace_id,
        data=data,
        text="ok",
        control_value_id=control_value_id)
    sending_queue.put(trace)


def send_report(
        incoming_value,
        sending_queue,
        network,
        device,
        value,
        state,
        trace_id=None):
    """
    Send a report message to the server.

    Sends a report message to the server containing data about the state of the
    specific value in a specific device. Puts a report message into
    sending_queue.

    Args:
        incoming_value: Value received from the server.
        sending_queue: The queue requests are being added to.
        network: Network the device is on.
        device: Device the value is on.
        value: Value the state belongs to.
        state: State of the device.
        trace_id: Trace ID used to create a URL for debugging (default: {None})

    """
    report = send_data.SendData(
        send_data.SEND_REPORT,
        data=str(incoming_value),
        network_id=network.uuid,
        device_id=device.uuid,
        value_id=value.uuid,
        state_id=state.uuid,
        trace_id=trace_id
    )
    sending_queue.put(report)


def get_control(incoming_value, network, device, value, state, trace_id=None):
    """
    Send a control message to the server.

    Sends a control message to the server containing data about a change in the
    state of the value of the specific device.

    Args:
        incoming_value: Value received from the server.
        network: Network the device is on.
        device: Device the value is on.
        value: Value the state belongs to.
        state: State of the device.
        trace_id: Trace ID used to create a URL for debugging.
            (default: {None})

    Returns:
        The result of the control request.

    """
    control = send_data.SendData(
        send_data.SEND_CONTROL,
        data=str(incoming_value),
        network_id=network.uuid,
        device_id=device.uuid,
        value_id=value.uuid,
        state_id=state.uuid,
        trace_id=trace_id
    )
    return control


class Handlers:
    """
    Message data handlers.

    Processes incoming and outgoing PUT, POST and GET messages.
    """

    def __init__(self, instance):
        """
        Initialize the Handler class.

        Create an instance of the Handlers class to handle incoming and
        outgoing communication with the server.

        Args:
            instance: Reference to the object instance class.

        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())
        self.instance = instance

    def handle_incoming_put(
            self,
            control_id,
            incoming_value,
            sending_queue,
            trace_id
    ):
        """
        Process an incoming PUT request.

        Handles an incoming PUT request and changes the specific value to
        the incoming value.

        Args:
            control_id: UUID of the Control state.
            incoming_value: Value received from the server.
            sending_queue: The queue requests are being added to.
            trace_id: Trace ID used to create a URL for debugging.
                (default: {None})

        Returns:
            True when successfully handling the request, False otherwise.

        """
        random_id = None
        if trace_id:
            network_n = self.instance.network_cl.name
            random_int = random.randint(1, 25000)
            random_id = "{}".format(network_n + random_int)
        for device in self.instance.device_list:
            for value in device.value_list:
                if value.control_state is not None:
                    if value.control_state.uuid == control_id:
                        if value.send_control(data_value=incoming_value):
                            send_trace(
                                sending_queue,
                                value.uuid,
                                trace_id,
                                incoming_value,
                                control_value_id=random_id
                            )
                            control = get_control(
                                incoming_value,
                                self.instance.network_cl,
                                device,
                                value,
                                value.control_state,
                                random_id
                            )
                            sending_queue.put(control)
                            return True
                        else:
                            return False
                else:
                    self.wapp_log.warning("Value is read only.")

        msg = "Unhandled put {} : {}".format(control_id, incoming_value)
        self.wapp_log.warning(msg)
        return False

    def handle_incoming_get(self, report_id, sending_queue, trace_id):
        """
        Process an incoming GET request.

        Handles an incoming GET request and sends the current value to the
        server.

        Args:
            report_id: UUID of the Report state.
            sending_queue: The queue requests are being added to.
            trace_id: Trace ID used to create a URL for debugging.
                (default: {None})


        Returns:
            True when successfully handling the request, False otherwise.

        """
        random_id = None
        if trace_id:
            random_int = str(random.randint(1, 25000))
            network_name = self.instance.network_cl.name

            random_id = "{}{}".format(network_name, random_int)

        for device in self.instance.device_list:
            for value in device.value_list:
                if value.report_state is not None:
                    if report_id.endswith(value.report_state.uuid):
                        current_value = value.data_value
                        send_trace(
                            sending_queue,
                            value.uuid,
                            trace_id,
                            current_value,
                            control_value_id=random_id
                        )
                        value.last_update_of_report = value.get_now()
                        send_report(
                            current_value,
                            sending_queue,
                            self.instance.network_cl,
                            device,
                            value,
                            value.report_state,
                            random_id
                        )
                        if value.callback is not None:
                            value.callback(value, 'report')
                        return True
                    else:
                        self.wapp_log.warning("Value ID mismatch.")
                else:
                    self.wapp_log.warning("Value is write only.")

        self.wapp_log.warning("Unhandled get {}".format(report_id))
        return False



from setuptools import find_packages, setup

with open("README.md", "r") as file:
    long_description = file.read()

setup(
    name="wappsto",
    version="1.0.0",
    author="Seluxit A/S",
    author_email="support@seluxit.com",
    license="Apache-2.0",
    description="Python Package to connect to wappsto.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/wappsto/wappsto-python",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"
    ],
    packages=find_packages(),
    tests_require=[
        'pytest',
    ],
)



"""
The raspberry object saving module.

Saves the running instance of the raspberry objects to a file.
"""
import os
import logging
from . import encoder


class SaveObjects:
    """
    The SaveObjects class that handles saving the running state to a file.

    Saves and loads the running instance of the raspberrypi objects to a JSON
    file in the saving folder.
    """

    def __init__(self, path_to_calling_file):
        """
        Initialize the WappstoEncoder class.

        Initializes a reference to the WappstoEncoder class so that it can be
        used to encode the current running instance into a JSON file.
        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())
        self.wappsto_encoder = encoder.WappstoEncoder()
        self.path_to_calling_file = path_to_calling_file

    def save_instance(self, instance):
        """
        Save the current instance as a JSON file.

        Encodes the current instance into a JSON file that can later be passed
        to the Instantiator class in order to parse it and create a new
        instance.

        Args:
            instance: Reference to the instance object that holds the network,
            device, value and state instances.

        Raises:
            OSError: There was an error opening or writing to the file.

        """
        # Creates the path.
        path = os.path.join(self.path_to_calling_file, 'saved_instances/')
        if not os.path.exists(path):
            os.mkdir(path)

        encoded_string = str(self.wappsto_encoder.encode(instance))
        encoded_string = str(encoded_string).replace("\'", "\\\"")
        encoded_string = str(encoded_string).replace(" ", "")
        encoded_string = '{"data":"' + encoded_string + '"}'

        network_id = instance.network_cl.uuid

        path_open = os.path.join(path, '{}.json'.format(network_id))

        with open(path_open, "w+") as network_file:
            network_file.write(encoded_string)

        msg = "Saved {} to {}".format(encoded_string, network_file)
        self.wapp_log.debug(msg)

    def load_instance(self):
        """
        Load the instance from the JSON file.

        Passes the path of the latest saved JSON file to allow parsing it.

        Returns:
            Path string to load file from.

        Raises:
            ValueError: An error occured finding files in the saved_instances
            directory.

        """
        path = os.path.join(self.path_to_calling_file, 'saved_instances/')
        files = os.listdir(path)
        file_paths = []
        latest_file = None
        for file_name in files:
            file_paths.append(os.path.join(path, file_name))
        try:
            latest_file = str(max(file_paths, key=os.path.getctime))
        except ValueError as ve:
            self.wapp_log.error("Exception in finding latest file: {}"
                                .format(ve))
            raise ve
        self.wapp_log.debug('Latest file: {}'.format(latest_file))

        return latest_file



"""
The __init__ method for the wappsto module.

Stores the Wappsto class functionality.
"""

import os
import json
import time
import logging
import inspect
from .connection import seluxit_rpc
from .connection import communication
from .connection.network_classes.errors import wappsto_errors
from .object_instantiation import status
from .object_instantiation import instantiate
from .object_instantiation import save_objects
from .user_guide.guider import Guider
RETRY_LIMIT = 5


class Wappsto:
    """
    The main package classs.

    Establishes a connection to the wappsto server, initializes the classes,
    starts a sending/receiving thread.
    """

    def __init__(self, json_file_name=None, load_from_state_file=False,
                 save_init=False):
        # TODO(Dimitar): Come up with a better description.
        """
        Initialize wappsto class.

        Creates a wappsto object which methods are exposed for users to use.
        These are: starting and stopping connection with a server and
        retrieving devices. These methods are enough for users to make
        operations on their network.

        Args:
            json_file_name: name of a json file containing all information
                about a network (default: {None})
            load_from_state_files: Defines if the data should be loaded from
                saved files (default: {False})
            save_init: Determines whether or not save json data
                (default: {False})

        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())

        # TODO(Dimitar): Comment on this later.
        stack = inspect.stack()[1][1]
        self.path_to_calling_file = os.path.dirname(os.path.abspath(stack))

        self.connecting = True
        self.rpc = seluxit_rpc.SeluxitRpc(save_init)
        self.socket = None
        self.receive_thread = None
        self.send_thread = None
        self.status = status.Status()
        self.object_saver = save_objects.SaveObjects(self.path_to_calling_file)
        self.guide = self.check_if_guide(self.path_to_calling_file)
        # Instantiate the objects from JSON
        try:
            self.instance = instantiate.Instantiator(
                json_file_name=json_file_name,
                load_from_state_file=load_from_state_file,
                status=self.status,
                guide=self.guide,
                path_to_calling_file=self.path_to_calling_file
            )
        # When the file fails to open a FileNotFoundError is raised and
        # the service is stopped
        except FileNotFoundError as fnfe:
            self.wapp_log.error("Failed to open file: {}".format(fnfe))
            self.stop(False)
            raise fnfe

    def check_if_guide(self, file_path):
        """
        Guide handler method.

        Checks if this is the user's first time using the package, then starts
        up a guide module to explain the various steps to the user.

        Returns:
            A boolean flag to designate if this is the first time launching the
            program.

        """
        path = os.path.join(file_path, 'config.txt')
        with open(path, 'r+') as json_file:
            data = json.load(json_file)
            guide = data['guide']
            if guide == "True":
                data['guide'] = "False"
                json_file.seek(0)
                json.dump(data, json_file)
                json_file.truncate()
                return True
            return False

    def get_status(self):
        """
        Wappsto status.

        Retrieves the status class instance which is set to different statuses
        throughout the program execution.

        Returns:
            A reference to the status object instance.

        """
        return self.status

    def get_device(self, name):
        """
        Device reference.

        Finds the device with a specific name attribute based on a given
        string

        Args:
            name: String containing the name attribute to search for.

        Returns:
            A reference to the device object instance.
            device

        Raises:
            DeviceNotFoundException: Device {name} not found in {instance}.

        """
        try:
            for device in self.instance.device_list:
                if name == device.name:
                    return device
            else:
                msg = "Device {} not found in {}".format(name, self.instance)
                raise wappsto_errors.DeviceNotFoundException(msg)
        except wappsto_errors.DeviceNotFoundException as device_not_found:
            msg = "Device not found {}".format(device_not_found)
            self.wapp_log.error(msg, exc_info=True)
            self.stop(False)
            raise device_not_found

    def start(self, address="wappsto.com", port=21005):
        """
        Start the server connection.

        Creates a socket object instance and passes the address and port to it.

        Args:
            address: Address to connect the service to.
                (default: {"qa.wappsto.com"})
            port: Port to connect the address to. (default: {31006})

        Raises:
            ServerConnectionException: "Unable to connect to the server.

        """
        self.status.set_status(status.STARTING)
        if self.guide:
            Guider.starting_server()
        # TODO(Dimitar): Change try except wrap to not encompass whole block.
        try:
            # Instance the socket class.
            reconnect_attempt_counter = 0
            self.socket = communication.ClientSocket(
                rpc=self.rpc,
                instance=self.instance,
                address=address,
                port=port,
                path_to_calling_file=self.path_to_calling_file
            )
            self.status.set_status(status.CONNECTING)
            if self.guide:
                Guider.connecting_to_server()

            # Attempts to connect to the server.
            self.connected = self.socket.connect()
            if self.connected:
                self.status.set_status(status.CONNECTED)
                if self.guide:
                    Guider.connected_to_server()
            # If it cannot connect it begins attempting the connection
            # until the retry limit is reached
            while (not self.connected
                    and reconnect_attempt_counter
                    < RETRY_LIMIT):
                self.status.set_status(status.RECONNECTING)
                if self.guide:
                    Guider.reconnecting_to_server()
                msg = "Cannot connect, attempting again in 5 seconds ..."
                self.wapp_log.info(msg)
                time.sleep(5)
                reconnect_attempt_counter += 1
                self.connected = self.socket.connect()

            # If the connection is not established,
            # a custom ServerConnectionException is raised.
            if reconnect_attempt_counter == RETRY_LIMIT:
                msg = ("Unable to connect to the server[IP: {}, Port: {}]"
                       .format(self.socket.address, self.socket.port)
                       )
                raise wappsto_errors.ServerConnectionException(msg)
        except wappsto_errors.ServerConnectionException as ce:
            msg = "Could not connect: {}".format(ce)
            self.wapp_log.error(msg, exc_info=True)
            self.stop(False)
            raise ce

        self.status.set_status(status.INITIALIZING)
        if self.guide:
            Guider.initializing()
        # Initializes the network, and all the subsequent devices, values and
        # states.
        # TODO(Dimitar): Change from generic Exception
        try:
            self.socket.initialize_all()
        except Exception as e:
            self.wapp_log.error("Error initializing: {}".format(e))
            self.stop(False)
            raise e

        self.status.set_status(status.STARTING_THREADS)
        if self.guide:
            Guider.starting_threads()
        # Starts the sending & receiving threads.
        try:
            self.receive_thread = self.socket.receiving_thread.start()
            self.send_thread = self.socket.sending_thread.start()
        except Exception as e:
            msg = "Error starting threads: {}".format(e)
            self.wapp_log.error(msg, exc_info=True)
            self.stop(False)
            raise e

        self.status.set_status(status.RUNNING)
        if self.guide:
            Guider.running()

    def stop(self, save=True):
        """
        Stop the Wappsto service.

        Closes the connection socket to the server and, if the boolean is set
        to True it saves an instance of the current state.

        Args:
            save: Flag to determine whether runtime instances should be saved.
                (default: {True})

        """
        # TODO(Dimitar): Add Exception checking if necessary.
        self.connecting = False
        self.status.set_status(status.DISCONNECTING)
        if self.guide:
            Guider.disconnecting()
        # Closes the socket connection, if one is established.
        if self.socket:
            self.socket.close()
        if save:
            self.object_saver.save_instance(self.instance)
        self.wapp_log.info("Exiting...")



"""
The device module.

Stores attributes for the device instance and handles device-related
methods.
"""
import logging
from .errors import wappsto_errors


class Device:
    """
    Device instance class.

    Stores attributes for the device instance and handles device-related
    methods.
    """

    def __init__(
        self,
        parent_network,
        uuid,
        name,
        product,
        protocol,
        serial_number,
        version,
        manufacturer,
        communication,
        description
    ):
        """
        Initialize the Device class.

        Initializes an object of device class by passing required parameters.

        Args:
            parent_network: Reference to an object of network class
            uuid: Unique identifier of a device
            name: Name of a device
            product: Defines what a product is
            protocol: Protocol device communicates by
            serial_number: Serial number of a device
            version: Version of a device
            manufacturer: Defines who made a product
            communication: The way a device communicates
            description: Description of a device

        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())
        self.parent_network = parent_network
        self.uuid = uuid
        self.name = name
        self.product = product
        self.serial_number = serial_number
        self.version = version
        self.manufacturer = manufacturer
        self.communication = communication
        self.protocol = protocol
        self.description = description
        self.value_list = []
        msg = "Device {} Debug: \n {}".format(name, str(self.__dict__))
        self.wapp_log.debug(msg)

    def get_parent_network(self):
        """
        Retrieve parent network reference.

        Gets a reference to the network that owns this device.

        Returns:
            Reference to instance of Network class that owns this Device.

        """
        return self.parent_network

    def add_value(self, value):
        """
        Append value reference to the value list.

        Adds a value reference to the list of value references
        ("owned" values).

        Args:
            value: Reference to instance of Value class.

        """
        self.value_list.append(value)
        self.wapp_log.debug("Value {} has been added.".format(value))

    def get_value(self, value_name):
        """
        Retrieve child value reference.

        Gets a child Value class reference by the given name string.

        Args:
            value_name: String to search by in the list of values.

        Returns:
            Reference to instance of Value class.

        """
        for value in self.value_list:
            if value_name in list(value.__dict__.values()):
                return value
            else:
                msg = "Value {} not found".format(value_name)
                wappsto_errors.ValueNotFoundException(msg)



"""
The server communication module.

Stores the server communication functionality for Wappsto as well as the
sending and receiving threads.
"""

import os
import socket
import threading
import time
import json
import queue
import ssl
# REPLACED request WITH NATIVE MODULE
import urllib.request as request
import logging
from . import send_data
from . import initialize
from . import handlers

t_url = 'https://tracer.iot.seluxit.com/trace?id={}&parent={}&name={}&status={}'  # noqa: E501


class ClientSocket:
    """
    The ClientSocket class that handles sending and receiving.

    Stores the sending and receiving threads, certificates, connection
    information as well as the related methods for handling communication
    between the client and the server.
    """

    def __init__(self, rpc, instance, address, port, path_to_calling_file):
        """
        Create a client socket.

        Creates a socket instance for the given address and port. Hhandles
        transfer of data from the instance attributes and methods to the
        specified server. Connection to the server is based on the specified
        address and port.

        Args:
            rpc: Sending/receiving queue processing instance.
            instance: Instance of network, devices, values and states.
            address: Server address.
            port: Server port.

        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())
        self.network = instance.network_cl
        self.instance = instance
        self.path_to_calling_file = path_to_calling_file
        self.ssl_server_cert = os.path.join(path_to_calling_file,
                                            "certificates/ca.crt")
        self.ssl_client_cert = os.path.join(path_to_calling_file,
                                            "certificates/client.crt")
        self.ssl_key = os.path.join(path_to_calling_file,
                                    "certificates/client.key")
        self.address = address
        self.port = port
        self.receiving_thread = threading.Thread(target=self.receive_thread)
        self.receiving_thread.setDaemon(True)
        self.connected = False
        self.sending_queue = queue.Queue(maxsize=0)
        self.sending_thread = threading.Thread(target=self.send_thread)
        self.sending_thread.setDaemon(True)
        self.rpc = rpc
        self.handlers = handlers.Handlers(self.instance)
        self.initialize_code = initialize.Initialize(self.rpc)
        self.packet_awaiting_confirm = {}
        self.add_trace_to_report_list = {}
        self.lock_await = threading.Lock()
        self.set_sockets()
        self.set_report_states()

    def set_report_states(self):
        """
        Set the reference to the queue and connection.

        Provides value classes with a referece to the queue and socket
        instances to enable report sending.
        """
        for device in self.instance.device_list:
            for value in device.value_list:
                value.rpc = self.rpc
                value.conn = self

    def set_sockets(self):
        """
        Create socket to communicate with server.

        Creates a socket instance and sets the options for communication.
        Passes the socket to the ssl_wrap method
        """
        self.my_raw_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.my_raw_socket.setsockopt(
            socket.SOL_SOCKET,
            socket.SO_KEEPALIVE,
            1
        )
        if (hasattr(socket, "TCP_KEEPIDLE")
                and hasattr(socket, "TCP_KEEPINTVL")
                and hasattr(socket, "TCP_KEEPCNT")):
            # After 5 idle minutes, start sending keepalives every 1 minutes.
            # Drop connection after 2 failed keepalives
            self.my_raw_socket.setsockopt(
                socket.SOL_TCP,
                socket.TCP_KEEPIDLE,
                5 * 60
            )
            self.my_raw_socket.setsockopt(
                socket.IPPROTO_TCP,
                socket.TCP_KEEPIDLE,
                5 * 60
            )
            self.my_raw_socket.setsockopt(
                socket.IPPROTO_TCP,
                socket.TCP_KEEPINTVL,
                60
            )
            self.my_raw_socket.setsockopt(
                socket.IPPROTO_TCP,
                socket.TCP_KEEPCNT,
                2
            )
            # self.my_raw_socket.setsockopt(
            #     socket.IPPROTO_TCP,
            #     socket.TCP_USER_TIMEOUT,
            #     30000
            # )
        self.my_socket = self.ssl_wrap()

    def ssl_wrap(self):
        """
        Wrap socket.

        Wraps the socket using the SSL protocol with the provided server and
        client certificates, as well as the SSL key.

        Returns:
        An SSL wrapped socket with the attached certificates/keys.

        """
        return ssl.wrap_socket(
            self.my_raw_socket,
            ca_certs=self.ssl_server_cert,
            certfile=self.ssl_client_cert,
            keyfile=self.ssl_key,
            cert_reqs=ssl.CERT_REQUIRED
        )

    def connect(self):
        """
        Connect to the server.

        Attempts a connection to the server on the provided addres and port.

        Returns:
            A connection flag to denote if the connection was successful or
            not.

        """
        self.connected = False
        try:
            self.my_socket.settimeout(10)
            self.my_socket.connect((self.address, self.port))
            self.connected = True
            self.my_socket.settimeout(None)
            self.wapp_log.info('Connected to server!')
            return True

        except Exception as e:
            self.wapp_log.error("Failed to connect: {}".format(e))
            return False

    def initialize_all(self):
        """
        Call initialize_all method in initialize_code module.

        Initializes the object instances on the sending/receiving queue.
        """
        self.initialize_code.initialize_all(self, self.instance)

    def add_id_to_confirm_list(self, data):
        """
        Add the message ID to the confirm list.

        Adds the ID of the decoded JSON message to the list of confirmed
        packets. Uses locks to ensure atomicity.

        Args:
            data: JSON communication message data.

        """
        self.lock_await.acquire()
        decoded = json.loads(data.decode('utf-8'))
        self.packet_awaiting_confirm[decoded.get('id')] = data
        self.lock_await.release()

    def remove_id_from_confirm_list(self, _id):
        """
        Remove the ID from the confirm list.

        Removes the ID of the decoded JSON message from the list of confirmed
        packets. Uses locks to ensure atomicity.

        Args:
            _id: ID to remove from the confirm list.

        """
        self.lock_await.acquire()
        if _id in self.packet_awaiting_confirm:
            del self.packet_awaiting_confirm[_id]
        self.lock_await.release()

    def incoming_control(self, data):
        """
        Incoming data handler.

        Sends the data from incoming control messages to the appropriate
        handler method.

        Args:
            data: JSON communication message data.

        Returns:
            Results of the incoming control handling.

        """
        return_id = data.get('id')
        try:
            control_id = data.get('params').get('data').get('meta').get('id')
        except AttributeError as e:
            error_str = 'Error received incorrect format in put'
            return self.handle_incoming_error(data, e, error_str, return_id)
        self.wapp_log.info("Control Request from control id: " + control_id)

        try:
            local_data = data.get('params').get('data').get('data')
        except AttributeError:
            error = 'Error received incorrect format in put, data missing'
            self.send_error(error, return_id)
            return
        try:
            trace_id = data.get('params').get('meta').get('trace')
            self.wapp_log.info("Control found trace id: " + trace_id)
        except AttributeError:
            # ignore
            trace_id = None

        if self.handlers.handle_incoming_put(
                control_id,
                local_data,
                self.sending_queue,
                trace_id
        ):
            self.send_success_reply(return_id)
        else:
            error = 'Invalid value range or non-existing ID'
            self.send_error(error, return_id)

    def handle_incoming_error(self, data, error_str, return_id):
        """
        Handle errors on the receive thread.

        Receives an error message and delivers it to the appropriate method.

        Args:
            data: JSON communication message data.
            error_str: Error contents.
            return_id: ID of the error message.

        """
        self.wapp_log.info(data)
        self.wapp_log.error(error_str, exc_info=True)
        self.send_error(error_str, return_id)
        return

    def send_success_reply(self, return_id):
        """
        Handle successful replies on the receive thread.

        Send back a success message based on the message return ID.

        Args:
            return_id: ID of the success message.

        """
        success_reply = send_data.SendData(
            send_data.SEND_SUCCESS,
            rpc_id=return_id
        )
        self.sending_queue.put(success_reply)

    def send_error(self, error_str, return_id):
        """
        Send error message.

        Sends back an error message based on the message return ID.

        Args:
            error_str: Error message contents.
            return_id: ID of the error message.

        """
        error_reply = send_data.SendData(
            send_data.SEND_FAILED,
            rpc_id=return_id,
            text=error_str
        )
        self.sending_queue.put(error_reply)

    def incoming_report_request(self, data):
        """
        Incoming data handler.

        Sends the data from incoming report messages to the appropriate handler
        method.

        Args:
            data: JSON communication message data.

        Returns:
            Results of the incoming report handling.

        """
        return_id = data.get('id')
        try:
            get_url_id = data.get('params').get('url')
        except AttributeError as e:
            error_str = 'Error received incorrect format in get'
            msg = "Report Request from url ID: {}".format(get_url_id)
            self.wapp_log.info(msg)
            return self.handle_incoming_error(data, e, error_str, return_id)

        try:
            trace_id = data.get('params').get('meta').get('trace')
            self.wapp_log.info("Report GET found trace id: {}"
                               .format(trace_id))
        except AttributeError:
            trace_id = None

        if self.handlers.handle_incoming_get(
                get_url_id,
                self.sending_queue,
                trace_id
        ):
            self.send_success_reply(return_id)
        else:
            error = 'Non-existing ID for get'
            self.send_error(error, return_id)

    def receive_thread(self):
        """
        Create the receive thread.

        Starts a socket to receive data and decode it. Based on the type of
        request, directs the decoded data to the appropriate methods.
        """
        self.wapp_log.debug("ReceiveThread Started!")
        while True:
            try:
                data = self.my_socket.recv(2000)
                decoded = json.loads(data.decode('utf-8'))
                decoded_id = decoded.get('id')
                try:
                    self.wapp_log.debug('Raw received Json: {}'
                                        .format(decoded))
                    if decoded.get('method', False) == 'PUT':
                        self.incoming_control(decoded)

                    elif decoded.get('method', False) == 'GET':
                        self.incoming_report_request(decoded)

                    elif decoded.get('error', False):
                        decoded_error = decoded.get('error')
                        msg = "Error: {}".format(decoded_error.get('message'))
                        self.wapp_log.error(msg)
                        self.remove_id_from_confirm_list(decoded_id)

                    elif decoded.get('result', False):
                        msg = "Successful reply for id {}".format(decoded_id)
                        self.wapp_log.debug(msg)
                        self.remove_id_from_confirm_list(decoded_id)

                    else:
                        self.wapp_log.info("Unhandled method")
                        error_str = 'Unknown method'
                        self.send_error(error_str, decoded_id)

                except ValueError:
                    self.wapp_log.info("Value error")
                    self.wapp_log.info(data)
                    error_str = 'Value error'
                    self.send_error(error_str, decoded_id)

            except ValueError:
                if data == b'':
                    self.reconnect()
                else:
                    self.wapp_log.info("Value error")
                    self.wapp_log.info(data)
                    error_str = 'Value error'
                    self.send_error(error_str, decoded_id)

            except ConnectionResetError as e:
                msg = "Received Reset: {}".format(e)
                self.wapp_log.error(msg, exc_info=True)
                self.reconnect()

            except OSError as oe:
                msg = "Received OS Error: {}".format(oe)
                self.wapp_log.error(msg, exc_info=True)
                self.reconnect()

    def reconnect(self):
        """
        Attempt to reconnect.

        Reconnection attemps in the instance of a connection being interrupted.
        """
        self.wapp_log.info("Server Disconnect")
        self.connected = False
        while not self.connected:
            self.wapp_log.info("Trying to reconnect in 5 seconds")
            time.sleep(5)
            self.my_socket.close()
            self.my_raw_socket.close()
            try:
                self.set_sockets()
                self.my_socket.connect((self.address, self.port))
                self.wapp_log.info("Reconnected")
                self.connected = True
                reconnect_reply = send_data.SendData(send_data.SEND_RECONNECT)
                self.sending_queue.put(reconnect_reply)
            except Exception as e:
                msg = "Failed to reconnect {}".format(e)
                self.wapp_log.error(msg, exc_info=True)

    def send_data(self, data):
        """
        Send JSON data.

        Sends the encoded JSON message through the socket.

        Args:
            data: JSON communication message data.

        """
        self.wapp_log.debug('Raw Send Json: {}'.format(data))
        self.my_socket.send(data)

    def send_thread(self):
        """
        Create a send thread.

        Retrieves packages from the sending queue to
        send data.
        """
        self.wapp_log.debug("SendingThread Started!")

        while True:
            package = self.sending_queue.get()
            if self.connected:
                if package.msg_id == send_data.SEND_SUCCESS:
                    self.send_success(package)

                elif package.msg_id == send_data.SEND_REPORT:
                    self.send_report(package)

                elif package.msg_id == send_data.SEND_FAILED:
                    self.send_failed(package)

                elif package.msg_id == send_data.SEND_RECONNECT:
                    self.send_reconnect()

                elif package.msg_id == send_data.SEND_CONTROL:
                    self.send_control(package)

                elif package.msg_id == send_data.SEND_TRACE:
                    self.send_trace(package)

                else:
                    self.wapp_log.info("Unhandled send")

            self.sending_queue.task_done()

    def send_trace(self, package):
        """
        Send data trace.

        Provides a trace URL for debugging purposes.

        Args:
            package: Sending queue item.

        """
        if package.control_value_id:
            control_value_id = package.control_value_id
            self.add_trace_to_report_list[control_value_id] = package.trace_id

        attempt = str(t_url).format(
            package.trace_id,
            package.parent,
            package.data,
            package.text
        )

        trace_req = request.urlopen(attempt)
        msg = "Sending tracer https message {} response {}".format(
            attempt,
            trace_req.getcode()
        )
        self.wapp_log.info(msg)

    def send_control(self, package):
        """
        Send data handler.

        Sends the data from outgoing control messages to the appropriate
        handler method.

        Args:
            package: Sending queue item.

        """
        self.wapp_log.info("Sending control message")
        try:
            local_data = self.rpc.get_rpc_state(
                package.data,
                package.network_id,
                package.device_id,
                package.value_id,
                package.state_id,
                'control',
                trace_id=package.trace_id
            )
            self.add_id_to_confirm_list(local_data)
            self.send_data(local_data)
        except OSError as e:
            self.connected = False
            msg = "Error sending control: {}".format(e)
            self.wapp_log.error(msg, exc_info=True)

    def send_reconnect(self):
        """
        Send a reconnect attempt.

        Sends a request to attempt to reconnect to the server.
        """
        self.wapp_log.info("Sending reconnect data")
        try:
            rpc_network = self.rpc.get_rpc_network(
                self.network.uuid,
                self.network.name,
                put=False
            )
            self.send_data(rpc_network)
            for element in self.packet_awaiting_confirm:
                self.send_data(self.packet_awaiting_confirm[element])
        except OSError as e:
            self.connected = False
            msg = "Error sending reconnect: {}".format(e)
            self.wapp_log.error(msg, exc_info=True)

    def send_failed(self, package):
        """
        Send a fail message.

        Sends a message to notify about a sending failure.

        Args:
            package: Sending queue item.

        """
        self.wapp_log.info("Sending Error")
        rpc_fail_response = self.rpc.get_rpc_fail_response(
            package.rpc_id,
            package.text
        )
        self.wapp_log.info(rpc_fail_response)
        try:
            self.send_data(rpc_fail_response)
        except OSError as e:
            self.connected = False
            msg = "Error sending failed response: {}".format(e)
            self.wapp_log.error(msg, exc_info=True)

    def send_report(self, package):
        """
        Send a report.

        Sends a report message from the package.

        Args:
            package: A sending queue item.

        """
        try:
            if not package.trace_id:
                if package.value_id in self.add_trace_to_report_list.keys():
                    package.trace_id = (
                        self.add_trace_to_report_list.pop(package.value_id)
                    )
            local_data = self.rpc.get_rpc_state(
                package.data,
                package.network_id,
                package.device_id,
                package.value_id,
                package.state_id,
                'report',
                trace_id=package.trace_id
            )
            self.add_id_to_confirm_list(local_data)
            self.send_data(local_data)
            decoded = json.loads(local_data.decode('utf-8'))
            data_decoded = decoded.get('params').get('data').get('data')
            self.wapp_log.info('Sending report value: {}'.format(data_decoded))
        except OSError as e:
            self.connected = False
            msg = "Error sending report: {}".format(e)
            self.wapp_log.error(msg, exc_info=True)

    def send_success(self, package):
        """
        Send a success message.

        Sends a message to notify of a successful.

        Args:
            package: A sending queue item.

        """
        try:
            rpc_success_response = self.rpc.get_rpc_success_response(
                package.rpc_id
            )
            self.send_data(rpc_success_response)
            self.wapp_log.info("Sending Successful")

        except OSError as e:
            self.connected = False
            msg = "Error sending response: {}".format(e)
            self.wapp_log.error(msg, exc_info=True)

    def close(self):
        """
        Close the connection.

        Closes the socket object connection.
        """
        self.wapp_log.info("Closing connection...")
        self.my_socket.close()
        self.my_raw_socket.close()

    def init_ok(self):
        """
        Return an ok from connecting.

        Return a message to signify that the connection was started
        successfully.

        Returns:
            A positive response.

        """
        data = self.my_socket.recv(4000)
        decoded = json.loads(data.decode('utf-8'))

        self.wapp_log.debug('Raw init response {}'.format(data))

        if data == b'':
            self.wapp_log.info("Server disconnected")
            return False

        if decoded.get('result', False):
            decoded_result = decoded.get('result')
            self.wapp_log.debug('Init result {}'.format(decoded_result))
            return True

        elif decoded.get('error'):
            msg = decoded.get('error').get('message')
            self.wapp_log.error(msg, exc_info=True)
            return False

        else:
            self.wapp_log.error(decoded, exc_info=True)
            return False
